import numpy as np
from typing import Union
from streamad.base import BaseDetector
from streamad.process import ZScoreCalibrator, TDigestCalibrator


class ConfiguredDetector:
    def __init__(self, detector: BaseDetector, calibrator: Union[ZScoreCalibrator, TDigestCalibrator] = None,
                 feature_index: int = None):
        self.detector = detector
        self.calibrator = calibrator
        self.feature_index = feature_index

    def fit_score(self, x):
        score = self.detector.fit_score(x) if self.feature_index is None or len(x) == 1 else self.detector.fit_score(
            np.array([x[self.feature_index]], np.float64))
        return self.calibrator.normalize(score) if self.calibrator else score

    def __str__(self):
        return self.detector.data_type + (
            ' (feature index: ' + str(self.feature_index) + ')' if self.feature_index is not None else '') + ' ' + type(
            self.detector).__name__ + ' with ' + type(self.calibrator).__name__



import pandas as pd
from streamad.evaluate import PointAwareMetircs
from streamad.util import CustomDS, plot
from ..model import LearningDetector


def evaluate(model: LearningDetector, train_ds: CustomDS = None, test_ds: CustomDS = None, init_window=200):
    assert (train_ds is not None or test_ds is not None, 'At least one dataset must be provided!')
    # initializing detector
    init_data = (train_ds if train_ds is not None else test_ds).data[:init_window]
    model.predict(init_data)

    # initializing metrics
    metrics = PointAwareMetircs()
    results = []

    if train_ds is not None:
        # train data evaluation before training
        scores = model.predict(train_ds.data)
        precision, recall, F1 = metrics.evaluate(train_ds.label, scores)
        results.append(['untrained', precision, recall, F1])

        # training model
        model.fit(train_ds.data, train_ds.label)

        # train data evaluation after training
        scores = model.predict(train_ds.data)
        precision, recall, F1 = metrics.evaluate(train_ds.label, scores)
        results.append(['trained', precision, recall, F1])

    if test_ds is not None:
        # test data evaluation after training
        scores = model.predict(test_ds.data)
        precision, recall, F1 = metrics.evaluate(test_ds.label, scores)
        results.append(['test', precision, recall, F1])

    return pd.DataFrame.from_records(results, columns=['name', 'precision', 'recall', 'F1']).set_index('name')



class LearningDetectorData:
    def __init__(self, features_count: int, coefs: dict, bias: float):
        self.features_count = features_count
        self.coefs = coefs
        self.bias = bias

    def __repr__(self):
        return "{cls_name}({features_count}, {coefs}, {bias})".format(cls_name=type(self).__name__,
                                                                      features_count=self.features_count,
                                                                      coefs=self.coefs, bias=self.bias)



import numpy as np
from sklearn.neural_network import MLPClassifier


class CustomMLPClassifier(MLPClassifier):
    def __init__(self, inputs_count: int, **kwargs):
        self.inputs_count = inputs_count
        super().__init__(**kwargs)
        super()._validate_input(np.array([[0] * inputs_count] * 2, np.float64), np.array([0, 1]), False, True)
        layer_units = [inputs_count] + list(self.hidden_layer_sizes) + [1]
        super()._initialize(np.zeros((1, 1)), layer_units, np.float64)

    def _init_coef(self, fan_in, fan_out, dtype):
        coef_init = np.full((fan_in, fan_out), 2 / (fan_in * fan_out), dtype=dtype)
        intercept_init = np.zeros(fan_out, dtype=dtype)
        return coef_init, intercept_init



import inspect
import numpy as np
import streamad.model as streamad_models
from streamad.process import ZScoreCalibrator
from ..data import LearningDetectorData
from ..base import ConfiguredDetector
from .classifier import CustomMLPClassifier
from typing import Callable


class LearningDetector:
    def __init__(self, features_count: int, logging_func: Callable[[int], None] = None):
        self.features_count = features_count
        self.detectors = []
        self.logging_func = logging_func

        for name, detector in inspect.getmembers(streamad_models, inspect.isclass):
            if name == 'RandomDetector':
                continue
            current_detector = detector()
            if current_detector.data_type == 'multivariate':
                self.detectors.append(ConfiguredDetector(current_detector, ZScoreCalibrator(sigma=3)))
            else:
                for feature_index in range(features_count):
                    self.detectors.append(ConfiguredDetector(detector(), ZScoreCalibrator(sigma=3), feature_index))

        self.classifier = CustomMLPClassifier(inputs_count=len(self.detectors), solver='lbfgs', hidden_layer_sizes=(),
                                              warm_start=True)

    def get_data(self):
        coefs = dict()
        for i in range(len(self.detectors)):
            coefs[str(self.detectors[i])] = self.classifier.coefs_[0][i][0]
        return LearningDetectorData(self.features_count, coefs, self.classifier.intercepts_[0][0])

    @classmethod
    def create_from_data(cls, data: LearningDetectorData):
        learning_detector = cls(data.features_count)
        for i in range(len(learning_detector.detectors)):
            detector_str = str(learning_detector.detectors[i])
            learning_detector.classifier.coefs_[0][i][0] = data.coefs.get(detector_str, 0.0)
        learning_detector.classifier.intercepts_[0][0] = data.bias
        return learning_detector

    def _get_scores(self, xs):
        if self.logging_func is None:
            return np.array([[detector.fit_score(x) for detector in self.detectors] for x in xs], np.float32)
        else:
            scores = []
            for i in range(len(xs)):
                scores.append([detector.fit_score(xs[i]) for detector in self.detectors])
                self.logging_func(i)
            return np.array(scores, np.float32)

    def fit(self, xs, ys):
        scores = self._get_scores(xs)
        ys = ys[~np.isnan(scores).any(axis=1)]
        scores = scores[~np.isnan(scores).any(axis=1)]
        self.classifier.fit(scores, ys)

    def predict(self, xs):
        scores = self._get_scores(xs)
        scores[np.isnan(scores)] = 0
        return self.classifier.predict(scores)



from os import path
from setuptools import setup, find_packages
from io import open

this_directory = path.abspath(path.dirname(__file__))

with open(path.join(this_directory, "mtehis/version.py")) as f:
    exec(f.read())


setup(
    name="mtehis",
    version=__version__,
    description="Maintenance Tool for Evolving Hybrid Intelligent Systems: Core Package.",
    long_description=open("./README.md").read(),
    long_description_content_type="text/markdown",
    author="Oleksandr Pokhylenko",
    author_email="pokhilenko.alex@gmail.com",
    packages=find_packages(exclude=["test"]),
    platforms=["all"],
    install_requires=["streamad==0.3.0"],
    include_package_data=True,
    url="https://github.com/MT-EHIS/core",
    setup_requires=["setuptools==58.2.0"],
    classifiers=[
        "Intended Audience :: Developers",
        "Intended Audience :: Education",
        "Intended Audience :: Science/Research",
        "Intended Audience :: Information Technology",
        "Programming Language :: Python :: 3.9",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
    ],
)
